import 'dart:ffi';

import 'package:dart_native/dart_native.dart';
import 'package:dart_native/src/darwin/runtime/internal/nsobject_lifecycle.dart';

final id nil = id(nullptr);

typedef Finalizer = void Function();

/// Stands for `NSObject` in iOS and macOS.
///
/// The root class of most Objective-C class hierarchies, from which subclasses inherit a basic interface to the runtime system and the ability to behave as Objective-C objects.
class NSObject extends id {
  Finalizer? _finalizer;
  Finalizer? get finalizer => _finalizer;
  set finalizer(Finalizer? f) {
    removeFinalizerForObject(this);
    _finalizer = f;
    addFinalizerForObject(this);
  }

  NSObject([Class? isa]) : super(_new(isa)) {
    bindLifecycleForObject(this);
  }

  /// Before calling [fromPointer], MAKE SURE the [ptr] for object exists.
  /// If [ptr] was already freed, you would get a crash!
  NSObject.fromPointer(Pointer<Void> ptr) : super(ptr) {
    bindLifecycleForObject(this);
  }

  NSObject init() {
    return perform(SEL('init'));
  }

  NSObject copy() {
    NSObject result = perform(SEL('copy'));
    return NSObject.fromPointer(result.autorelease().pointer);
  }

  NSObject mutableCopy() {
    NSObject result = perform(SEL('mutableCopy'));
    return NSObject.fromPointer(result.autorelease().pointer);
  }

  NSObject autorelease() {
    return perform(SEL('autorelease'));
  }

  static Pointer<Void> _new(Class? isa) {
    isa ??= Class('NSObject');
    Pointer<Void> resultPtr = isa.perform(SEL('new'), decodeRetVal: false);
    return msgSend(resultPtr, SEL('autorelease'), decodeRetVal: false);
  }
}

Pointer<Void> alloc(Class? isa) {
  isa ??= Class('NSObject');
  Pointer<Void> resultPtr = isa.perform(SEL('alloc'), decodeRetVal: false);
  return msgSend(resultPtr, SEL('autorelease'), decodeRetVal: false);
}

typedef ConvertorFromPointer = dynamic Function(Pointer<Void> ptr);

/// Register a function for converting a Dart object from a [Pointer].
///
/// Example for [NSString]:
///
/// ```dart
/// registerTypeConvertor('NSString', (ptr) {
///     return NSString.fromPointer(ptr);
/// });
/// ```
/// The example above can be generated by applying `@native` annotation on Dart
/// wrapper class. See [NSString].
void registerTypeConvertor(String type, ConvertorFromPointer convertor) {
  if (_convertorCache[type] == null) {
    _convertorCache[type] = convertor;
  }
}

/// Convert [arg] to its custom type, which is annotated with `@native()`.
dynamic objcInstanceFromPointer(Pointer<Void> arg, String? type) {
  if (arg == nullptr) {
    return arg;
  }
  // delete '?' for null-safety
  if (type != null) {
    if (type.endsWith('?')) {
      type = type.substring(0, type.length - 1);
    }
  } else {
    /// Retrive class name from native.
    var object = NSObject.fromPointer(arg);
    type = object.isa?.name;
  }
  ConvertorFromPointer? convertor = _convertorCache[type];
  if (convertor != null) {
    return convertor(arg);
  }
  return NSObject.fromPointer(arg);
}

Map<String, ConvertorFromPointer> _convertorCache = {};
